from dataclasses import dataclass
from typing import List, Tuple

import equinox as eqx
import orbax.checkpoint as ocp
import orbax.checkpoint.future as ocp_future
from etils import epath


class EquinoxCheckpointHandler(ocp.AsyncCheckpointHandler):
    """Handler for equinox pytree."""

    def __init__(self):
        pass

    def save(
        self,
        directory: epath.Path,
        args: "EquinoxSave",
        **kwargs,
    ):
        eqx.tree_serialise_leaves(
            str(directory / "model.eqx"), (args.model, args.state)
        )

    async def _save(self, directory, args, **kwargs):
        self.save(directory, args, **kwargs)

    def restore(
        self,
        directory: epath.Path,
        args: "EquinoxRestore",
        **kwargs,
    ) -> Tuple[eqx.Module, eqx.nn.State]:
        model, state = eqx.tree_deserialise_leaves(
            str(directory / "model.eqx"), (args.model, args.state)
        )
        if args.state._state == {}:
            return model
        return model, state

    async def async_save(
        self,
        directory: epath.Path,
        args: "EquinoxSave",
        **kwargs,
    ) -> List[ocp_future.CommitFutureAwaitingContractedSignals]:
        return [
            ocp_future.CommitFutureAwaitingContractedSignals(
                self._save(directory, args),
                name="equinox_save",
            )
        ]

    def close(self):
        pass


@ocp.args.register_with_handler(EquinoxCheckpointHandler, for_save=True)
@dataclass
class EquinoxSave(ocp.args.CheckpointArgs):
    """Arguments for saving an equinox model."""

    model: eqx.Module
    state: eqx.nn.State = eqx.nn.State(None)


@ocp.args.register_with_handler(EquinoxCheckpointHandler, for_restore=True)
@dataclass
class EquinoxRestore(ocp.args.CheckpointArgs):
    """Arguments for restoring equinox model."""

    model: eqx.Module
    state: eqx.nn.State = eqx.nn.State(None)


CheckpointManager = ocp.CheckpointManager
CheckpointManagerOptions = ocp.CheckpointManagerOptions
Checkpointer = ocp.Checkpointer
AsyncCheckpointer = ocp.AsyncCheckpointer

args = ocp.args



import re
from pathlib import Path
from typing import Iterable, List, Tuple


def recursive_folder_parsing(
    root_dir: Path, parsers: List[re.Pattern]
) -> Iterable[Tuple[Path, dict]]:
    """Iteratively go through the subdirectories of root_dir.
    The ones that match the parser tree are returned along with their parsed arguments
    in a flattened way.
    The structure to be parsed is specified via a list of expected patterns, organized
    according to the folder levels. The first pattern specifies the folders to select,
    the second patterns the subfolders, and so on.

    :param root_dir: root directory.
    :param parsers: ordered list of regular expression patterns.
        Every patterns specified the folders that are considered at every level.
    :return: an iterable of (directory, parameters) tuples.
    """
    for subdir in root_dir.glob("*"):
        match = parsers[0].match(subdir.name)
        if match is None:
            # logging.info("Recursive parsing: skipped due to not parsing %s" % subdir)
            continue

        if len(parsers) == 1:
            yield subdir, match.groupdict()
        else:
            for subsubdir, matchdict in recursive_folder_parsing(subdir, parsers[1:]):
                yield subsubdir, {**match.groupdict(), **matchdict}



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg


def fig2im(fig, tight_layout=True):
    canvas = FigureCanvasAgg(fig)
    if tight_layout:
        fig.tight_layout(pad=0.0)
    canvas.draw()
    image = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
    image = image.reshape(canvas.get_width_height()[::-1] + (4,))  # Shape: (H, W, 4)
    plt.close(fig)
    return image



import logging
from pathlib import Path
from typing import Literal, Optional, OrderedDict, Union

from rich.logging import RichHandler

LOG_LEVEL = Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class Logger(logging.Logger):
    name: Optional[str] = None
    path: Union[str, Path, None] = None

    @classmethod
    def getLogger(
        cls,
        name: str,
        path: Union[str, Path, None] = None,
        level: LOG_LEVEL = "INFO",
    ) -> "Logger":
        logger = Logger(logging.getLogger(name))
        logger.name = name
        logger.path = path
        if not logger.handlers:
            # terminal handler
            xterm_handler = RichHandler(rich_tracebacks=True, markup=True)
            xterm_handler.setLevel(level)
            xterm_handler.setFormatter(
                logging.Formatter(fmt="%(message)s", datefmt="[%c]")
            )
            logger.addHandler(xterm_handler)

            # file handler
            if path is not None:
                log_path = Path(path)
                if log_path.is_dir():
                    log_path = log_path / f"{name}.log"
                file_handler = logging.FileHandler(log_path)
                file_handler.setLevel(level)
                file_handler.setFormatter(
                    logging.Formatter(
                        fmt="[%(asctime)s] [%(levelname)s] %(message)s",
                        datefmt="%Y-%m-%d %H:%M:%SZ",
                    ),
                )
                logger.addHandler(file_handler)
            logger.propagate = False
        logger.setLevel(level)
        return logger

    @classmethod
    def num2pstr(cls, x: int | float, fmt: str = ".3f") -> str:
        """Convert number to pretty string with metric suffixes"""
        if x < 0:
            prefix = "-"
            x = -x
        else:
            prefix = ""
        suffix_table = OrderedDict(
            {
                "1e18": "E",
                "1e15": "P",
                "1e12": "T",
                "1e9": "G",
                "1e6": "M",
                "1e3": "K",
                "1e0": "",
                "1e-3": "m",
                "1e-6": "µ",
                "1e-9": "n",
                "1e-12": "p",
                "1e-15": "f",
                "1e-18": "a",
            }
        )
        suffix = ""
        for k, v in suffix_table.items():
            if x >= float(k):
                x /= max(k, 1e-9)
                suffix = v
                break
        return f"{prefix}{x:{fmt}}{suffix}"

    @classmethod
    def pstr2num(cls, s: str) -> float:
        """Convert pretty string with metric suffixes to number (might be lossy depending on fmt used!)"""
        suffix_table = {
            "E": 1e18,
            "P": 1e15,
            "T": 1e12,
            "G": 1e9,
            "M": 1e6,
            "K": 1e3,
            "": 1e0,
            "m": 1e-3,
            "µ": 1e-6,
            "u": 1e-6,  # allow 'u' as 'µ'
            "n": 1e-9,
            "p": 1e-12,
            "f": 1e-15,
            "a": 1e-18,
        }
        s = s.strip()
        if not s:
            raise ValueError("Empty string")
        # Find the position where the numeric part ends
        num_part = ""
        suffix_part = ""
        for i, char in enumerate(s):
            if char.isdigit() or char in ".-+":
                num_part += char
            else:
                suffix_part = s[i:].strip()
                break
        if not num_part:
            raise ValueError(f"No numeric part found in '{s}'")
        try:
            num_value = float(num_part)
        except ValueError as e:
            raise ValueError(f"Invalid numeric part '{num_part}' in '{s}'") from e
        if suffix_part not in suffix_table:
            raise ValueError(f"Unknown suffix '{suffix_part}' in '{s}'")
        return num_value * suffix_table[suffix_part]



from typing import Tuple, Union

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array


class StreamNormalizer(eqx.Module):
    mean_index: eqx.nn.StateIndex
    var_index: eqx.nn.StateIndex
    count_index: eqx.nn.StateIndex

    def __init__(self, shape: Union[int, Tuple[int, ...]]):
        if isinstance(shape, int):
            shape = (shape,)
        self.mean_index = eqx.nn.StateIndex(jnp.zeros(shape))
        self.var_index = eqx.nn.StateIndex(jnp.ones(shape))
        self.count_index = eqx.nn.StateIndex(0)

    def __call__(self, x: Array, state: eqx.nn.State) -> Array:
        """center and scale"""
        return (x - state.get(self.mean_index)) / jnp.sqrt(
            state.get(self.var_index) + 1e-8
        )

    def center(self, x: Array, state: eqx.nn.State) -> Array:
        return x - state.get(self.mean_index)

    def update(self, x: Array, state: eqx.nn.State) -> eqx.nn.State:
        """NOTE: this method operates on a batch of vectors"""
        batch_mean = jnp.mean(x, axis=0)
        batch_var = jnp.var(x, axis=0)
        batch_count = x.shape[0]

        mean = state.get(self.mean_index)
        var = state.get(self.var_index)
        count = state.get(self.count_index)

        delta = batch_mean - mean
        tot_count = count + batch_count

        new_mean = mean + delta * batch_count / tot_count
        m_a = var * count
        m_b = batch_var * batch_count
        M2 = m_a + m_b + jnp.square(delta) * batch_count * count / tot_count
        new_var = M2 / tot_count
        new_count = tot_count

        new_state = (
            state.set(self.mean_index, new_mean)
            .set(self.var_index, new_var)
            .set(self.count_index, new_count)
        )
        return new_state


class EMAPercScaler(eqx.Module):
    """EMA scale by difference of percentiles (Dreamer-v3 tricks for Policy Gradient (REINFORCE))"""

    low_index: eqx.nn.StateIndex
    high_index: eqx.nn.StateIndex

    low_p: float = eqx.field(static=True)
    high_p: float = eqx.field(static=True)
    tau: float = eqx.field(static=True, default=0.01)

    def __init__(
        self,
        shape: Union[int, Tuple[int, ...]],
        low_p: float = 0.05,
        high_p: float = 0.95,
        tau: float = 0.01,
    ):
        assert 0 <= low_p < high_p <= 1, (
            "Percentiles must be in [0, 1] and low_p < high_p"
        )
        if isinstance(shape, int):
            shape = (shape,)

        self.low_p = low_p
        self.high_p = high_p
        self.tau = tau

        self.low_index = eqx.nn.StateIndex(jnp.full(shape, jnp.nan))
        self.high_index = eqx.nn.StateIndex(jnp.full(shape, jnp.nan))

    def __call__(self, x: Array, state: eqx.nn.State) -> Array:
        """center and scale"""
        low_p = state.get(self.low_index)
        high_p = state.get(self.high_index)
        scale = jnp.nan_to_num(high_p - low_p, nan=1.0)
        return x / jnp.maximum(scale, 1.0)

    def update(self, x: Array, state: eqx.nn.State) -> eqx.nn.State:
        """NOTE: this method operates on a batch of vectors"""
        batch_low = jnp.percentile(x, self.low_p, axis=0)
        batch_high = jnp.percentile(x, self.high_p, axis=0)

        prev_low = state.get(self.low_index)
        prev_high = state.get(self.high_index)

        new_low = jnp.where(
            jnp.isnan(prev_low),
            batch_low,
            prev_low + self.tau * (batch_low - prev_low),
        )
        new_high = jnp.where(
            jnp.isnan(prev_high),
            batch_high,
            prev_high + self.tau * (batch_high - prev_high),
        )
        new_state = state.set(self.low_index, new_low).set(self.high_index, new_high)
        return new_state



"""Commonly occurring patterns for regular expression parsing."""

FLOAT_PATTERN = r"[-+]?(?:\d*\.\d+|\d+)"
INT_PATTERN = r"\d+"
BOOL_PATTERN = r"[tT]rue|[fF]alse|[tT]|[fF]"
BOOL_WITH_NUMERIC_PATTERN = r"[tT]rue|[fF]alse|[tT]|[fF]|0|1"
LITERAL_PATTERN = r"[a-zA-Z]+"
# Numerical format, including exponential format.
NUMERIC_PATTERN = r"[-+]? (?: (?: \d* \. \d+ ) | (?: \d+ \.? ) )(?: [Ee] [+-]? \d+ ) ?"



import matplotlib as mpl
import matplotlib.pyplot as plt
from adjustText import adjust_text

from .configs import use
from .utils import fig2im

__all__ = ["mpl", "plt", "use", "fig2im", "adjust_text"]
